package cz.mendelu.dnaAnalyser.utils.analyse.service;

import cz.mendelu.dnaAnalyser.utils.analyse.model.Analyse;
import cz.mendelu.dnaAnalyser.utils.model.Identifiable;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.Configuration;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.TransactionFactory;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.jdbc.datasource.DriverManagerDataSource;

import javax.sql.DataSource;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.sql.*;
import java.util.UUID;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static java.text.MessageFormat.format;

@Slf4j
public abstract class AbstractDataSessionService implements DataSessionService {

    private static final String TABLE_NAME = "TABLE_NAME";
    private static final Pattern RE_JDBC_FILE = Pattern.compile("^jdbc:h2:file:(.*?);");
    private static final String DB_H2_FILE = ".mv.db";

    @Value("${analyse.datasource.url}")
    protected String datasourceUrl;
    @Value("${analyse.datasource.username}")
    protected String datasourceUsername;
    @Value("${analyse.datasource.password}")
    protected String datasourcePassword;
    @Value("${analyse.datasource.driverClassName}")
    protected String datasourceDriverClassName;

    protected SqlSessionFactoryBuilder sessionFactoryBuilder = new SqlSessionFactoryBuilder();

    @Override
    public <A extends Identifiable> SqlSession openDataSession(A analyse) {
        String analyseId = analyse.getId().toString();
        String jdbcUrl = datasourceUrl.replace("analyse-id", analyseId);

        // Create Datasource and tables
        DataSource dataSource = createDataSource(jdbcUrl);
        if (notExistsDatabaseFile(jdbcUrl)) {
            createTables(jdbcUrl, dataSource);
        }

        // Mybatis configuration
        TransactionFactory transactionFactory = new JdbcTransactionFactory();
        Environment environment = new Environment(analyseId, transactionFactory, dataSource);
        Configuration configuration = new Configuration(environment);
        configuration.setLazyLoadingEnabled(true);
        onConfiguration(configuration);

        SqlSessionFactory factory = sessionFactoryBuilder.build(configuration);
        return factory.openSession();
    }

    @Override
    public <A extends Identifiable> Connection openConnection(A analyse) throws SQLException {
        String analyseId = analyse.getId().toString();
        String jdbcUrl = datasourceUrl.replace("analyse-id", analyseId);

        // Create Datasource and tables
        DataSource dataSource = createDataSource(jdbcUrl);
        if (notExistsDatabaseFile(jdbcUrl)) {
            createTables(jdbcUrl, dataSource);
        }
        return dataSource.getConnection();
    }


    @Override
    public Connection openConnection(UUID uuid) throws SQLException {
        String analyseId = uuid.toString();
        String jdbcUrl = datasourceUrl.replace("analyse-id", analyseId);

        // Create Datasource and tables
        DataSource dataSource = createDataSource(jdbcUrl);
        if (notExistsDatabaseFile(jdbcUrl)) {
            createTables(jdbcUrl, dataSource);
        }
        return dataSource.getConnection();
    }

    protected abstract void onCreateTables(Statement statement) throws SQLException;

    protected abstract void onConfiguration(Configuration configuration);

    protected boolean notExistsDatabaseFile(String jdbcUrl) {
        Matcher matcher = RE_JDBC_FILE.matcher(jdbcUrl);
        if (matcher.find()) {
            Path file = Paths.get(matcher.group(1) + DB_H2_FILE);
            return Files.notExists(file);
        }
        throw new IllegalArgumentException(format("File not hound in {}", jdbcUrl));
    }

    protected DataSource createDataSource(String jdbcUrl){
        DriverManagerDataSource dataSource = new DriverManagerDataSource();
        dataSource.setDriverClassName(datasourceDriverClassName);
        dataSource.setUrl(jdbcUrl);
        dataSource.setUsername(datasourceUsername);
        dataSource.setPassword(datasourcePassword);
        log.info("Create datasource {}", dataSource.getUrl());
        return dataSource;
    }

    protected void createTables(String jdbcUrl, DataSource dataSource) {
        try (Connection connection = dataSource.getConnection()) {
            try (Statement statement = connection.createStatement()) {
                onCreateTables(statement);
            } catch (SQLException e) {
                log.error(format("Create tables: {0} failed.", jdbcUrl), e);
            }
        } catch (SQLException e) {
            log.error(format("Create tables: connection for {0} failed.", jdbcUrl), e);
        }
    }
}
